
#ifndef CSOUND_MEMALLOC_H
#define CSOUND_MEMALLOC_H

/* memalloc.h -- written by Istvan Varga, Feb 2003 */

#ifdef __cplusplus
extern "C" {
#endif

#define MEMALLOC_ERROR_RETURN   0x00000001
#define MEMALLOC_DISPOSABLE     0x00000002

/* initialise a heap for Csound instance 'csound' with the specified */
/* flags; flags is a bitwise OR of any of the following values:      */
/*   MEMALLOC_ERROR_RETURN                                           */
/*     alloc/free routines should return in case of a failure (with  */
/*     a NULL return value if applicable), rather than calling the   */
/*     csoundError() callback of 'csound' with an error message.     */
/*     If not set, alloc functions never return NULL (with the       */
/*     exception of realloc being called with a zero size).          */
/*   MEMALLOC_DISPOSABLE                                             */
/*     enable advanced management of allocated memory:               */
/*       csoundHeapDestroy() can free all the allocated memory       */
/*       improved error checking                                     */
/*     however, this makes the function calls slower, and increases  */
/*     memory usage by 32 to 48 bytes per allocated memory block     */
/* return value is a handle for the newly created heap, which may    */
/* be NULL if the MEMALLOC_ERROR_RETURN flag is set and an error     */
/* occured.                                                          */

void *csoundHeapCreate(void *csound, int flags);

/* Set error code with which CSOUNDAPI.Error is called as       */
/*   CSOUNDAPI.Error(csound, errCode, errMsg, handle)           */
/* for heap 'handle'.                                           */
/* The default error code is CSOUND_MEMORY.                     */

void csoundHeapSetErrorCode(void *handle, int errCode);

/* allocate a memory block for the heap identified by 'handle'  */
/* 'nbytes' is the number of bytes to allocate                  */
/* return value is a pointer to the new memory area             */

void *csoundMalloc(void *handle, size_t nbytes);

/* allocate a memory block for the heap identified by 'handle',    */
/* and clear it to zero                                            */
/* 'nmemb' and 'size' are the same as the parameters for calloc()  */
/* return value is a pointer to the new memory area                */

void *csoundCalloc(void *handle, size_t nmemb, size_t size);

/* free a memory block in heap identified by 'handle'  */
/* 'ptr' is a pointer returned by a previous call to   */
/* csoundMalloc(), csoundCalloc(), or csoundRealloc()  */
/* with the same handle.                               */

void csoundFree(void *handle, void *ptr);

/* re-allocate a memory block for the heap identified by 'handle'  */
/* 'ptr' and 'nbytes' are the same as the parameters for realloc() */
/* return value is a pointer to the new memory area                */

void *csoundRealloc(void *handle, void *ptr, size_t nbytes);

/* free all memory used by the specified heap         */
/* has no effect if called with handle = NULL         */
/* WARNING: the expected input parameter is a pointer */
/* to the variable storing the heap handle, and not   */
/* the handle itself; this variable is set to NULL    */
/* on return.                                         */

void csoundHeapDestroy(void **handle);

#ifdef __cplusplus
}
#endif

#endif          /* CSOUND_MEMALLOC_H */

