
/* memalloc.c -- written by Istvan Varga, Feb 2003 */

#define CS_H_NO_MACROS_FOR_GLOBALS  1

#include <cs.h>
#include "memalloc.h"

#define CS_MEMINFO_BLOCK_SIZE   256
#define CS_MEMINFO_BLOCK_MASK   255

#define CS_MEMHDR_MAGIC(x)      ((unsigned int) ((void*) (x))           \
                                 ^ (unsigned int) 0x78D93E41)

#ifdef exit
#undef exit
#endif

typedef struct CS_MEMINFO_S {
    void        *p;
    struct CS_MEMINFO_S
                *prv_free;
} CS_MEMINFO;

typedef struct CS_MEMINFO_BLOCK_S {
    CS_MEMINFO  allocs[CS_MEMINFO_BLOCK_SIZE];
    struct CS_MEMINFO_BLOCK_S
                *prv_block;
} CS_MEMINFO_BLOCK;

typedef struct CS_MEMHDR_S {
    unsigned int  magic;
    CS_MEMINFO    *p;
} CS_MEMHDR;

typedef struct CS_HEAP_S {
    int         alloc_count;
    int         max_alloc_count;
    int         flags;
    int         error_code;
    void        *csound;
    CS_MEMINFO  *free_alloc_stack;
    CS_MEMINFO_BLOCK
                *alloc_blocks;
} CS_HEAP;

#define memoryError(x,y,e,h)                                                  \
{                                                                             \
    if (!((y) & MEMALLOC_ERROR_RETURN)) {                                     \
      ((CSOUNDAPI*) (x))->Error((x), (e), "csound: not enough memory", (h));  \
    }                                                                         \
    return NULL;                                                              \
}

#define memoryErrorCheck(x,y,z,e,h) if ((x) == NULL) memoryError(y,z,e,h)

/* initialise a heap for Csound instance 'csound' with the specified */
/* flags; flags is a bitwise OR any of the following values:         */
/*   MEMALLOC_ERROR_RETURN                                           */
/*     alloc/free routines should return in case of a failure (with  */
/*     a NULL return value if applicable), rather than calling the   */
/*     Error() callback of 'csound' with an error message.           */
/*     If not set, alloc functions never return NULL (with the       */
/*     exception of realloc being called with a zero size).          */
/*   MEMALLOC_DISPOSABLE                                             */
/*     enable advanced management of allocated memory:               */
/*       csoundHeapDestroy() can free all the allocated memory       */
/*       improved error checking                                     */
/*     however, this makes the function calls slower, and increases  */
/*     memory usage by 32 to 48 bytes per allocated memory block     */
/* return value is a handle for the newly created heap, which may    */
/* be NULL if the MEMALLOC_ERROR_RETURN flag is set and an error     */
/* occured.                                                          */

void *csoundHeapCreate(void *csound, int flags)
{
    CS_HEAP *p;

    p = (CS_HEAP*) malloc(sizeof(CS_HEAP));
    memoryErrorCheck(p, csound, flags, CSOUND_MEMORY, NULL);

    p->alloc_count = 0;
    p->flags = flags;
    p->error_code = CSOUND_MEMORY;
    p->csound = csound;
    p->free_alloc_stack = NULL;
    if (flags & MEMALLOC_DISPOSABLE) {
      p->max_alloc_count = CS_MEMINFO_BLOCK_SIZE;
      p->alloc_blocks = (CS_MEMINFO_BLOCK*)
        calloc((size_t) 1, sizeof(CS_MEMINFO_BLOCK));
      if (p->alloc_blocks == NULL) {
        free(p);
        memoryError(csound, flags, CSOUND_MEMORY, NULL);
      }
      p->alloc_blocks->prv_block = NULL;
    }
    else {
      p->max_alloc_count = 0;
      p->alloc_blocks = NULL;
    }
    return (void*) p;
}

/* Set error code with which CSOUNDAPI.Error is called as   */
/*   CSOUNDAPI.Error(csound, errCode, errMsg, handle)       */
/* for heap 'handle'.                                       */
/* The default error code is CSOUND_MEMORY.                 */

void csoundHeapSetErrorCode(void *handle, int errCode)
{
    if (handle == NULL)
      return;
    ((CS_HEAP*) handle)->error_code = errCode;
}

static CS_MEMINFO *register_memory_alloc(CS_HEAP *h, void *p)
{
    CS_MEMINFO  *m;

    /* store magic number */
    ((CS_MEMHDR*) p)->magic = (unsigned int) CS_MEMHDR_MAGIC(h);
    /* allocate new info structure */
    if (h->free_alloc_stack != NULL) {
      /* pop from stack, */
      m = h->free_alloc_stack;
      h->free_alloc_stack = m->prv_free;
    }
    else {
      CS_MEMINFO_BLOCK  *b;
      /* or allocate new one */
      if (h->alloc_count >= h->max_alloc_count) {
        h->max_alloc_count += CS_MEMINFO_BLOCK_SIZE;
        b = (CS_MEMINFO_BLOCK*) calloc((size_t) 1, sizeof(CS_MEMINFO_BLOCK));
        memoryErrorCheck(b, h->csound, h->flags, h->error_code, (void*) h);
        b->prv_block = h->alloc_blocks;
        h->alloc_blocks = b;
      }
      m = &(h->alloc_blocks->allocs[h->alloc_count & CS_MEMINFO_BLOCK_MASK]);
      h->alloc_count++;
    }
    /* store pointer to info structure */
    ((CS_MEMHDR*) p)->p = m;
    /* update pointer so that it now points to the actual user memory area */
    p = (void*) ((uintptr_t) p + (intptr_t) 0x10);
    /* fill data fields in info structure */
    m->p = p;
    m->prv_free = NULL;

    return m;
}

static void *unregister_memory_alloc(CS_HEAP *h, void *p)
{
    CS_MEMINFO  *m;

    /* find header */
    p = (void*) ((uintptr_t) p - (intptr_t) 0x10);
    if (((CS_MEMHDR*) p)->magic != CS_MEMHDR_MAGIC(h))
      return NULL;      /* already freed or invalid address */
    /* no longer a valid header */
    ((CS_MEMHDR*) p)->magic = CS_MEMHDR_MAGIC(NULL);
    /* get pointer to info structure */
    m = ((CS_MEMHDR*) p)->p;
    /* push to free allocs stack */
    m->p = NULL;
    m->prv_free = h->free_alloc_stack;
    h->free_alloc_stack = m;

    return p;       /* succesfully unregistered */
}

/* allocate a memory block for the heap identified by 'handle'  */
/* 'nbytes' is the number of bytes to allocate                  */
/* return value is a pointer to the new memory area             */

void *csoundMalloc(void *handle, size_t nbytes)
{
    void    *p;
    CS_HEAP *h = (CS_HEAP*) handle;

    if (!(h->flags & MEMALLOC_DISPOSABLE)) {
      p = (void*) malloc(nbytes);
      memoryErrorCheck(p, h->csound, h->flags, h->error_code, (void*) h);
      return p;
    }
    else {
      CS_MEMINFO  *m;
      p = (void*) malloc(nbytes + (size_t) 0x10);
      memoryErrorCheck(p, h->csound, h->flags, h->error_code, (void*) h);
      m = register_memory_alloc(h, p);
      if (m == NULL) {
        free(p);        /* in case of error */
        memoryError(h->csound, h->flags, h->error_code, (void*) h);
      }
      return m->p;
    }
}

/* allocate a memory block for the heap identified by 'handle',    */
/* and clear it to zero                                            */
/* 'nmemb' and 'size' are the same as the parameters for calloc()  */
/* return value is a pointer to the new memory area                */

void *csoundCalloc(void *handle, size_t nmemb, size_t size)
{
    void    *p;
    CS_HEAP *h = (CS_HEAP*) handle;

    if (!(h->flags & MEMALLOC_DISPOSABLE)) {
      p = (void*) calloc(nmemb, size);
      memoryErrorCheck(p, h->csound, h->flags, h->error_code, (void*) h);
      return p;
    }
    else {
      CS_MEMINFO  *m;
      size_t      nbytes = nmemb * size;
      p = (void*) calloc(nbytes + (size_t) 0x10, (size_t) 1);
      memoryErrorCheck(p, h->csound, h->flags, h->error_code, (void*) h);
      m = register_memory_alloc(h, p);
      if (m == NULL) {
        free(p);        /* in case of error */
        memoryError(h->csound, h->flags, h->error_code, (void*) h);
      }
      return m->p;
    }
}

/* free a memory block in heap identified by 'handle'  */
/* 'ptr' is a pointer returned by a previous call to   */
/* csoundMalloc(), csoundCalloc(), or csoundRealloc()  */
/* with the same handle.                               */
/* return value is a pointer to the new memory area    */

void csoundFree(void *handle, void *ptr)
{
    CS_HEAP *h = (CS_HEAP*) handle;

    /* trivial case */
    if (ptr == NULL)
      return;

    if (!(h->flags & MEMALLOC_DISPOSABLE)) {
      free(ptr);
    }
    else {
      void        *p;
      /* find original (unaligned) pointer */
      p = unregister_memory_alloc(h, ptr);
      if (p == NULL) {
        if (!(h->flags & MEMALLOC_ERROR_RETURN)) {
          ((CSOUNDAPI*) h->csound)->Error(h->csound, h->error_code,
                                          "csoundFree(): invalid pointer",
                                          (void*) h);
        }
        return;
      }
      free(p);
    }
}

/* re-allocate a memory block for the heap identified by 'handle'  */
/* 'ptr' and 'nbytes' are the same as the parameters for realloc() */
/* return value is a pointer to the new memory area                */

void *csoundRealloc(void *handle, void *ptr, size_t nbytes)
{
    void    *p;
    CS_HEAP *h = (CS_HEAP*) handle;

    if (!(h->flags & MEMALLOC_DISPOSABLE)) {
      p = (void*) realloc(ptr, nbytes);
      if (nbytes != (size_t) 0) {   /* a NULL return is valid if size == 0 */
        memoryErrorCheck(p, h->csound, h->flags, h->error_code, (void*) h);
      }
      return p;
    }
    else {
      if (ptr != NULL) {
        /* original pointer is non-zero: standard realloc */
        p = unregister_memory_alloc(h, ptr);
        if (p == NULL) {
          if (!(h->flags & MEMALLOC_ERROR_RETURN)) {
            ((CSOUNDAPI*) h->csound)->Error(h->csound, h->error_code,
                                            "csoundRealloc(): invalid pointer",
                                            (void*) h);
          }
          return NULL;
        }
        if (nbytes != (size_t) 0) {
          CS_MEMINFO  *m;
          ptr = (void*) realloc(p, nbytes + (size_t) 0x10);
          memoryErrorCheck(ptr, h->csound, h->flags, h->error_code, (void*) h);
          m = register_memory_alloc(h, ptr);
          if (m == NULL) {
            free(ptr);
            memoryError(h->csound, h->flags, h->error_code, (void*) h);
          }
          return m->p;
        }
        else {
          free(p);      /* trivial case: size = 0 */
          return NULL;
        }
      }
      else {
        /* original pointer is zero: same as malloc */
        if (nbytes != (size_t) 0) {
          CS_MEMINFO  *m;
          p = (void*) malloc(nbytes + (size_t) 0x10);
          memoryErrorCheck(p, h->csound, h->flags, h->error_code, (void*) h);
          m = register_memory_alloc(h, p);
          if (m == NULL) {
            free(p);        /* in case of error */
            memoryError(h->csound, h->flags, h->error_code, (void*) h);
          }
          return m->p;
        }
        else {
          return NULL;  /* trivial case: size = 0 */
        }
      }
    }
}

/* free all memory used by the specified heap         */
/* has no effect if called with handle = NULL         */
/* WARNING: the expected input parameter is a pointer */
/* to the variable storing the heap handle, and not   */
/* the handle itself; this variable is set to NULL    */
/* on return.                                         */

void csoundHeapDestroy(void **handle)
{
    CS_HEAP *h;
    void    *p;
    int     i, j;

    if (handle == NULL)
      return;
    h = *((CS_HEAP**) handle);
    *((CS_HEAP**) handle) = NULL;
    if (h == NULL)
      return;
    i = h->alloc_count;
    while (--i >= 0) {
      j = i & CS_MEMINFO_BLOCK_MASK;
      if (h->alloc_blocks->allocs[j].p != NULL) {
        p = unregister_memory_alloc(h, h->alloc_blocks->allocs[j].p);
        if (p == NULL) {        /* assertion only */
          fprintf(stderr, "internal error in csoundHeapDestroy()\n");
          exit(-1);
        }
        free(p);
      }
      if (!j) {
        CS_MEMINFO_BLOCK  *b = h->alloc_blocks->prv_block;
        free(h->alloc_blocks);
        h->alloc_blocks = b;
      }
    }
    if (h->alloc_blocks != NULL)        /* may be necessary if   */
      free(h->alloc_blocks);            /* nothing was allocated */
    free(h);
}

