/*                                               RTAUDIO.C for Linux  */

/*  This module is included when RTAUDIO is defined at compile time.
    It provides an interface between Csound realtime record/play calls
    and the device-driver code that controls the actual hardware.
 */

#include <Csound/csdl.h>
#include <unistd.h>
#include <fcntl.h>

#undef ENGINE
#define ENGINE  ((CSOUNDAPI*) csound)
#undef Str
#define Str(x)  (((CSOUNDAPI*) csound)->LocalizeString(x))
#undef exit

/*
 *  Linux code based on examples in the "Hacker's Guide to VoxWare 2.4"
 *  and the file "experimental.txt" accompanying the source code
 *  for VoxWare 2.9 and later (provided with the Linux kernel source),
 *  both by Hannu Savolainen (hannu@voxware.pp.fi).
 *
 *  The code not only initializes the sound board and driver for the correct
 *  sample size, rate, and channels, but customizes the DMA buffers to
 *  match Csound's input or output buffer size.
 *
 *  A new option ('-V') was added to Csound to allow the user to set the
 *  master output volume on the soundcard.
 *
 *  Jonathan Mohr
 *  Augustana University College
 *  Camrose, Alberta, Canada  T4V 2R3
 *  mohrj@augustana.ab.ca
 *
 *  1995 Oct 17
 */

#include <sys/ioctl.h>
#include <sys/soundcard.h>

#define MIXER_NAME  "/dev/mixer"

static void setsndparms(void *csound,
                        int dspfd, int format, int nchanls, float sr,
                        unsigned bufsiz )
{
  int       parm, original;
  unsigned  frag_size;

  /* set sample size/format */
  switch ( format ) {
  case AE_UNCH:  /* unsigned char - standard Linux 8-bit format */
    parm = AFMT_U8;
    break;
  case AE_CHAR:  /* signed char. - probably not supported by Linux */
    parm = AFMT_S8;
    break;
  case AE_ULAW:
    parm = AFMT_MU_LAW;
    break;
  case AE_ALAW:
    parm = AFMT_A_LAW;
    break;
  case AE_SHORT:
#ifdef LINUX_BE
    parm = AFMT_S16_BE; /* Linux on SPARC is big-endian */
#else
    parm = AFMT_S16_LE; /* Linux on Intel x86 is little-endian */
#endif
    break;
  case AE_LONG: {
    ENGINE->Die(csound,
      Str("Linux sound driver does not support long integer samples"));
    return;
  }
  case AE_FLOAT: {
    ENGINE->Die(csound,
      Str("Linux sound driver does not support floating-point samples"));
    return;
  }
  default:
    /* Linux sound driver provides these names for modes not used by Csound */
    /* parm = AFMT_IMA_ADPCM; */
    /* parm = AFMT_U16_LE;    */
    /* parm = AFMT_U16_BE;    */
    ENGINE->Die(csound, Str("unknown sample format"));
    return;
  }
  original = parm;
  if (ioctl(dspfd, SOUND_PCM_WRITE_BITS, &parm) == -1) {
    ENGINE->Die(csound,
                Str("unable to set requested sample format on soundcard"));
    return;
  }
  if (parm != original) {
    ENGINE->Die(csound,
                Str("soundcard does not support the requested sample format"));
    return;
  }
  /* set number of channels (mono or stereo) */
  parm = nchanls;
  if (ioctl(dspfd, SOUND_PCM_WRITE_CHANNELS, &parm) == -1) {
    ENGINE->Die(csound, Str("unable to set mode (mono/stereo) on soundcard"));
    return;
  }
  if (parm != nchanls) {
    ENGINE->Die(csound,
      Str("DSP device does not support the requested mode (mono/stereo)"));
    return;
  }

  /* set the sample rate */
  parm = (int) sr;
  if (ioctl(dspfd, SOUND_PCM_WRITE_RATE, &parm) == -1) {
    ENGINE->Die(csound, Str("unable to set sample rate on soundcard"));
    return;
  }
  if (parm != (int) sr && (ENGINE->GetMessageLevel(csound) & 4)) {
    char msg[256];
    sprintf(msg, Str("Sample rate set to %d (instead of %d)\n"),
                 parm, (int) sr);
    ENGINE->Message(csound, msg);
  }

#ifndef __FreeBSD__
  /* set DMA buffer fragment size to Csound's output buffer size */
  parm = 0; frag_size = 1;
  /* find least power of 2 >= bufsiz */
  while ( frag_size < bufsiz ) {
    frag_size <<= 1;
    parm++;
  }
  parm |= 0x00020000;   /* Larry Troxler's idea */
  /*  parm |= 0x00ff0000; */    /* use max. number of buffer fragments */
  if (ioctl(dspfd, SNDCTL_DSP_SETFRAGMENT, &parm) == -1) {
    ENGINE->Die(csound,
                Str("failed while trying to set soundcard DMA buffer size"));
    return;
  }
  /* find out what buffer size the driver allocated */
  if (ioctl(dspfd, SNDCTL_DSP_GETBLKSIZE, &parm) == -1) {
    ENGINE->Die(csound,
                Str("failed while querying soundcard about buffer size"));
    return;
  }
  if (parm != (int) frag_size && (ENGINE->GetMessageLevel(csound) & 4)) {
    char msg[256];
    sprintf(msg,
            Str("Soundcard DMA buffer size set to %d bytes (instead of %d)\n"),
            parm, frag_size);
    ENGINE->Message(csound, msg);
  }
  else
    ENGINE->Message(csound, Str("hardware buffers set to %d bytes\n"), parm);
#endif
}

static void setvolume(void *csound, unsigned volume)
{
  int mixfd, parm;

  /* open mixer device for writing */
  if ( (mixfd = open(MIXER_NAME, O_WRONLY)) == -1 ) {
    ENGINE->Die(csound,
                Str("unable to open soundcard mixer for setting volume"));
    return;
  }
  /* set volume (left and right channels the same) */
  parm = (volume & 0xff) | ((volume & 0xff) << 8);
  if (ioctl(mixfd, SOUND_MIXER_WRITE_VOLUME, &parm) == -1) {
    ENGINE->Die(csound, Str("unable to set output volume on soundcard"));
    return;
  }
  /* close mixer device */
  if ( close(mixfd) == -1 ) {
    ENGINE->Die(csound, Str("error while closing sound mixer device"));
  }
}

#define DSP_NAME    "/dev/dsp"

static void*
recopen(void *csound, AUDIOPARMS *devParms)     /* open for audio input */
{
    int dspfd_in;

    /* Jonathan Mohr  1995 Oct 17 */
    /* open DSP device for reading */
    if ( (dspfd_in = open(DSP_NAME, O_RDONLY)) == -1 ) { /* IV - Nov 10 2002 */
      ENGINE->Die(csound, Str("unable to open soundcard for audio input"));
      return NULL;
    }
    /* initialize data format, channels, sample rate, and buffer size */
    setsndparms(csound, dspfd_in,
                devParms->sampleFormat,
                devParms->nChannels,
                devParms->sampleRate,
                (unsigned) (devParms->hwBufSize * devParms->sampleSize
                            * devParms->nChannels));
    return (void*) dspfd_in;
}

/* IV - Oct 26 2002: sr should be float */
static void*
playopen(void *csound, AUDIOPARMS *devParms)    /* open for audio output */
{
    int dspfd_out;
        /*  J. Mohr  1995 Oct 17 */
/* The following code not only opens the DSP device (soundcard and driver)
   for writing and initializes it for the proper sample size, rate, and
   channels, but allows the user to set the output volume.                  */
    /* open DSP device for writing */               /* IV - Nov 10 2002 */
    if ( (dspfd_out = open(DSP_NAME, O_WRONLY)) == -1 ) {
      ENGINE->Die(csound, Str("unable to open soundcard for audio output"));
      return NULL;
    }
    /* set sample size/format, rate, channels, and DMA buffer size */
    setsndparms(csound, dspfd_out,
                devParms->sampleFormat,
                devParms->nChannels,
                devParms->sampleRate,
                (unsigned) (devParms->hwBufSize * devParms->sampleSize
                            * devParms->nChannels));
    /* check if volume was specified as command line parameter */
    if (devParms->userParam1) {
      /* check range of value specified */
      if (devParms->userParam1 > 100 || devParms->userParam1 < 0) {
        close(dspfd_out);
        ENGINE->Die(csound, Str("Volume must be between 0 and 100"));
        return NULL;
      }
      setvolume(csound, devParms->userParam1);
    }
    return (void*) dspfd_out;
}

static int
rtrecord(void *audioInData, char *inbuf, int nbytes) /* get samples from ADC */
{
    /*  J. Mohr  1995 Oct 17 */
    if ( (nbytes = read(((int) audioInData), inbuf, nbytes)) == -1 ) {
      fprintf(stderr, "error while reading DSP device for audio input\n");
      exit(-1);
    }
    return(nbytes);
}

static void
rtplay(void *audioOutData, char *outbuf, int nbytes)  /* put samples to DAC  */
    /* N.B. This routine serves as a THROTTLE in Csound Realtime Performance, */
    /* delaying the actual writes and return until the hardware output buffer */
    /* passes a sample-specific THRESHOLD.  If the I/O BLOCKING functionality */
    /* is implemented ACCURATELY by the vendor-supplied audio-library write,  */
    /* that is sufficient.  Otherwise, requires some kind of IOCTL from here. */
    /* This functionality is IMPORTANT when other realtime I/O is occurring,  */
    /* such as when external MIDI data is being collected from a serial port. */
    /* Since Csound polls for MIDI input at the software synthesis K-rate     */
    /* (the resolution of all software-synthesized events), the user can      */
    /* eliminate MIDI jitter by requesting that both be made synchronous with */
    /* the above audio I/O blocks, i.e. by setting -b to some 1 or 2 K-prds.  */
{
    /*  J. Mohr  1995 Oct 17 */                     /* IV - Nov 10 2002 */
    if (write((int) audioOutData, outbuf, nbytes) < nbytes)
      fprintf(stderr, "/dev/dsp: could not write all bytes requested\n");
}

static void rtclose(void *audioInData, void *audioOutData)
                                /* close the I/O device entirely  */
{                               /* called only when both complete */
    /*  J. Mohr  1995 Oct 17 */
    if ((int) audioInData > 0 && (close((int) audioInData) == -1)) {
      fprintf(stderr, "unable to close DSP device\n");
      exit(-1);
    }
    if ((int) audioOutData > 0 && (close((int) audioOutData) == -1)) {
      fprintf(stderr, "unable to close DSP device\n");
      exit(-1);
    }
}

long opcode_count(void *csound)
{
    csound = csound;
    return 0L;
}

void *opcode_init(void *csound)
{
    fprintf(stderr, "OSS plugin by Jonathan Mohr\n");

    ENGINE->SetPlayopenCallback(csound, playopen);
    ENGINE->SetRtplayCallback(csound, rtplay);
    ENGINE->SetRecopenCallback(csound, recopen);
    ENGINE->SetRtrecordCallback(csound, rtrecord);
    ENGINE->SetRtcloseCallback(csound, rtclose);

    return NULL;
}

