/*                                               RTAUDIO.C for WIN32   */

/*  This module is included when RTAUDIO is defined at compile time.
    It provides an interface between Csound realtime record/play calls
    and the device-driver code that controls the actual hardware.
 */
/* RWD modified 15.7.99 - NT-friendly, removed format switch blocks,
   enabling m/c performance eliminated timer from playback code
 */
/* IV - Jan 11 2003: extensively revised code (many parts rewritten), and
   added these features:
     priority control (normal, high, critical)
     variable number of buffers (RWD version was hard-coded to 4)
     sleep time can be controlled for optimal performance
     method for preparing headers can be selected (also for performance)
     code cleaned up
   IV - Feb 22 2003: modified for use as plugin DLL
*/
#include <Csound/csdl.h>
#undef MYFLT
#include <time.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>

#undef ENGINE
#define ENGINE  ((CSOUNDAPI*) csound)
#undef Str
#define Str(x)  (((CSOUNDAPI*) csound)->LocalizeString(x))

/* IV - Jan 30 2003 */
#define MAXNUMBUF  (64)

typedef struct {
        CSOUNDAPI   *csound;
        HWAVEOUT    outdev;                 /* Device handle */
        int         numbuf_out;             /* number of buffers */
        LPSTR       wavbuff[MAXNUMBUF];     /* N buffers */
        WAVEHDR     wavhdr[MAXNUMBUF];      /* and associated headers */
        int         rwd_hdrs;               /* for tuning performance */
        int         rtwin_sleep;
        long        nrecs;
} WAVE_OUT_DATA;

typedef struct {
        CSOUNDAPI   *csound;
        HWAVEIN     indev;                  /* Device handle */
        int         numbuf_in;              /* number of buffers */
        LPSTR       wavInbuff[MAXNUMBUF];   /* N buffers */
        WAVEHDR     wavInhdr[MAXNUMBUF];    /* and associated headers */
        int         rwd_hdrs;               /* for tuning performance */
        int         rtwin_sleep;
        long        inrecs;
} WAVE_IN_DATA;

static  int     sched_priority = 2;     /* 0: critical, 1: high, 2: normal */

/* set scheduling priority */

static void set_priority_level(int n)
{
    n = (n < 0 ? 0 : (n > 2 ? 2 : n));
    if ((sched_priority < 2 && n < 2) ||
        (sched_priority == 2 && n == 2))        /* check if */
      return;                                   /* already set */
    sched_priority = n;         /* mark as already set */
    switch (sched_priority) {
    case 0:                     /* critical */
      SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS);
      SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);
      break;
    case 1:                     /* high */
      SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);
      SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL);
      break;
    default:                    /* normal */
      SetPriorityClass(GetCurrentProcess(), NORMAL_PRIORITY_CLASS);
      SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
      break;
    }
}

/* print error message for waveInOpen or waveOutOpen and exit */

static void rtopen_error(void *csound, int n)
{
    char *s = NULL, msg[256];
    switch (n) {
    case MMSYSERR_BADDEVICEID:
      s = Str("ID is out of range");
      break;
    case MMSYSERR_ALLOCATED:
      s = Str("already allocated");
      break;
    case MMSYSERR_NOMEM:
      s = Str("unable to allocate or lock memory");
      break;
    case WAVERR_BADFORMAT:
      s = Str("unsupported wave format");
      break;
    default:
      s = Str("???");
      break;
    }
    sprintf(msg, Str("Error code: %s\n"), s);
    ENGINE->Message(csound, msg);
    ENGINE->Die(csound, Str("Failed to open dac"));
}

static void rtopen_select_sample_format(void *csound,
                                        int fmt, float sr, int nchns,
                                        WAVEFORMATEX *wfx)
{
    switch (fmt) {
    case AE_UNCH:
      wfx->wBitsPerSample = 8;
      break;
    case AE_SHORT:
      wfx->wBitsPerSample = 16;
      break;
    case AE_LONG:
      wfx->wBitsPerSample = 32;
      break;
    default:
      ENGINE->Die(csound, Str("WAVE OUT unknown wave format"));
      break;
    }
    wfx->wFormatTag = WAVE_FORMAT_PCM;
    wfx->nChannels = nchns;
    wfx->nSamplesPerSec = (int) (sr + 0.5f);
    wfx->nAvgBytesPerSec =
      (int) (sr + 0.5f) * nchns * (wfx->wBitsPerSample >> 3);
    wfx->nBlockAlign = nchns * (wfx->wBitsPerSample >> 3);
}

static void*
recopen(void *csound, AUDIOPARMS *devParms)   /* open for audio input */
{
    int           ndev = (int) waveInGetNumDevs(), j, b, bufbytes;
    WAVEINCAPS    caps;
    WAVEFORMATEX  wavform;      /* RWD: I assume Win32s knows about this... */
    WAVE_IN_DATA  *audioInData;

    /* allocate instance data */
    audioInData =
      (WAVE_IN_DATA*) calloc((size_t) 1, (size_t) sizeof(WAVE_IN_DATA));
    if (audioInData == NULL) {
      ENGINE->Die(csound, "not enough memory");
      return NULL;
    }
    audioInData->csound = (CSOUNDAPI*) csound;
    audioInData->rwd_hdrs = (devParms->userParam1 ? 1 : 0);
    b = devParms->sleepTime;
    audioInData->rtwin_sleep = (b ? (b < 1 ? 1 : (b > 11 ? 11 : b)) - 1 : 1);
    /* select device */
    if (ndev < 1)
      ENGINE->Die(csound, Str("No sound input capabilities"));
    if (ndev > 1) {             /* We ought to allow used to choise but how? */
      ENGINE->Message(csound, Str("%d WAVE IN devices found\n"), ndev);
      for (j = 0; j < (int) ndev; j++) {
        waveInGetDevCaps(j, &caps, sizeof(caps));
        ENGINE->Message(csound, Str("WAVE IN device %d: %s\n"),
                                j, caps.szPname);
      }
      j = (devParms->deviceNumber < ndev ? devParms->deviceNumber : 0);
      ENGINE->Message(csound, Str("Selecting device %d\n"), j);
      ndev = j;
    }
    else ndev = 0;
    if (waveInGetDevCaps(ndev, &caps, sizeof(WAVEINCAPS)) != MMSYSERR_NOERROR)
      ENGINE->Die(csound, Str("Cannot get capabilities"));

    /* set sample format */
    rtopen_select_sample_format(csound, devParms->sampleFormat,
                                devParms->sampleRate, devParms->nChannels,
                                &wavform);

    /* open device */
    if ((b = (int) waveInOpen((LPHWAVEIN) &(audioInData->indev), ndev,
                              &wavform,
                              (DWORD) NULL,
                              (DWORD) NULL, CALLBACK_NULL))
        != MMSYSERR_NOERROR) {
      rtopen_error(csound, b);
      return NULL;
    }
    ENGINE->Message(csound, Str("--->WAVE IN DEV.#%d ENABLED  ( %s )\n"),
                            ndev, caps.szPname);

    /* set up buffers */
    bufbytes = devParms->swBufSize * (int) wavform.nBlockAlign;
    audioInData->numbuf_in = devParms->hwBufSize / devParms->swBufSize;
    for (b = 0; b < audioInData->numbuf_in; b++) {
      WAVEHDR *whdr = &(audioInData->wavInhdr[b]);
      audioInData->wavInbuff[b] = whdr->lpData =
        GlobalLock(GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, bufbytes));
      memset(audioInData->wavInbuff[b], 0, bufbytes);
      whdr->dwBufferLength = (DWORD) bufbytes;
      whdr->dwUser = 0;
      whdr->dwFlags = WHDR_DONE;
      whdr->dwLoops = 1;
      waveInPrepareHeader(audioInData->indev, whdr, sizeof(WAVEHDR));
    }
    ENGINE->Message(csound, Str("WAVE IN: using %d buffers of %d bytes\n"),
                            audioInData->numbuf_in, bufbytes);
    waveInStart(audioInData->indev);
    audioInData->inrecs = 0L;                        /* IV - Jan 15 2003 */
    /* set scheduling priority */
    set_priority_level(devParms->schedPriority ?
                       devParms->schedPriority - 1 : 2);

    return (void*) audioInData;
}

static void*
playopen(void *csound, AUDIOPARMS *devParms)   /* open for audio output */
{
    int           ndev = (int) waveOutGetNumDevs(), j, b, bufbytes;
    WAVEOUTCAPS   caps;
    WAVEFORMATEX  wavform;      /* RWD: I assume Win32s knows about this... */
    WAVE_OUT_DATA *audioOutData;

    /* allocate instance data */
    audioOutData =
      (WAVE_OUT_DATA*) calloc((size_t) 1, (size_t) sizeof(WAVE_OUT_DATA));
    if (audioOutData == NULL) {
      ENGINE->Die(csound, "not enough memory");
      return NULL;
    }
    audioOutData->csound = (CSOUNDAPI*) csound;
    audioOutData->rwd_hdrs = (devParms->userParam1 ? 1 : 0);
    b = devParms->sleepTime;
    audioOutData->rtwin_sleep = (b ? (b < 1 ? 1 : (b > 11 ? 11 : b)) - 1 : 1);
    /* select device */
    if (ndev < 1)
      ENGINE->Die(csound, Str("No sound capabilities"));
    if (ndev > 1) {             /* We ought to allow used to choise but how? */
      ENGINE->Message(csound, Str("%d WAVE OUT devices found\n"), ndev);
      for (j = 0; j < ndev; j++) {
        waveOutGetDevCaps(j, &caps, sizeof(caps));
        ENGINE->Message(csound, Str("WAVE OUT device %d: %s\n"),
                                j, caps.szPname);
      }
      j = (devParms->deviceNumber < ndev ? devParms->deviceNumber : 0);
      ENGINE->Message(csound, Str("Selecting device %d\n"), j);
      ndev = j;
    }
    else ndev = 0;
    if (waveOutGetDevCaps(ndev, &caps, sizeof(WAVEOUTCAPS)) != MMSYSERR_NOERROR)
      ENGINE->Die(csound, Str("Cannot get capabilities"));

    /* set sample format */
    rtopen_select_sample_format(csound, devParms->sampleFormat,
                                devParms->sampleRate, devParms->nChannels,
                                &wavform);

    /* open device */
    if ((b = (int) waveOutOpen((LPHWAVEOUT) &(audioOutData->outdev), ndev,
                               &wavform,
                               (DWORD) NULL,
                               (DWORD) NULL, CALLBACK_NULL))
        != MMSYSERR_NOERROR) {
      rtopen_error(csound, b);
      return NULL;
    }
    ENGINE->Message(csound, Str("--->WAVE OUT DEV.#%d ENABLED  ( %s )\n"),
                            ndev, caps.szPname);

    /* set up buffers */
    bufbytes = devParms->swBufSize * (int) wavform.nBlockAlign;
    audioOutData->numbuf_out = devParms->hwBufSize / devParms->swBufSize;
    for (b = 0; b < audioOutData->numbuf_out; b++) {
      WAVEHDR *whdr = &(audioOutData->wavhdr[b]);
      audioOutData->wavbuff[b] = audioOutData->wavhdr[b].lpData =
        GlobalLock(GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, bufbytes));
      memset(audioOutData->wavbuff[b], 0, bufbytes);
      whdr->dwBufferLength = bufbytes;
      whdr->dwUser = 0;
      whdr->dwFlags = WHDR_DONE;
      whdr->dwLoops = (audioOutData->rwd_hdrs ? 0 : 1);
    }
    ENGINE->Message(csound, Str("WAVE OUT: using %d buffers of %d bytes\n"),
                            audioOutData->numbuf_out, bufbytes);
    /* set scheduling priority */
    set_priority_level(devParms->schedPriority ?
                       devParms->schedPriority - 1 : 2);
    /* prepare headers now if enabled */
    if (!audioOutData->rwd_hdrs)
      for (b = 0; b < audioOutData->numbuf_out; b++)
        waveOutPrepareHeader(audioOutData->outdev,
                             &(audioOutData->wavhdr[b]), sizeof(WAVEHDR));

    return (void*) audioOutData;
}

static int rtrecord(WAVE_IN_DATA *audioInData, char *inbuf, int nbytes)
{                                               /* get samples from ADC */
    int   ndx = (int) audioInData->inrecs % audioInData->numbuf_in;
    int   prvndx = (ndx ? (ndx - 1) : (audioInData->numbuf_in - 1));
    volatile DWORD *dwflg = &(audioInData->wavInhdr[ndx].dwFlags);

    if (audioInData->inrecs >= audioInData->numbuf_in) {
      if (audioInData->rtwin_sleep)                     /* if input too fast */
        while ((*dwflg & WHDR_DONE) == 0) Sleep(audioInData->rtwin_sleep);
      else
        while ((*dwflg & WHDR_DONE) == 0);
    }
    /* copy data */
    memcpy(inbuf, audioInData->wavInbuff[ndx], nbytes);
    /* prepare new buffer */
    if (audioInData->rwd_hdrs) {
      waveInPrepareHeader(audioInData->indev,
                          &(audioInData->wavInhdr[prvndx]), sizeof(WAVEHDR));
      audioInData->wavInhdr[ndx].dwBufferLength = nbytes;       /* ??? */
    }
    waveInAddBuffer(audioInData->indev,
                    &(audioInData->wavInhdr[prvndx]), sizeof(WAVEHDR));
    /* next buffer */
    audioInData->inrecs++;
    return(nbytes);
}

static void rtplay(WAVE_OUT_DATA *audioOutData, char *outbuf, int nbytes)
                                                /* put samples to DAC  */
  /* N.B. This routine serves as a THROTTLE in Csound Realtime Performance, */
  /* delaying the actual writes and return until the hardware output buffer */
  /* passes a sample-specific THRESHOLD.  If the I/O BLOCKING functionality */
  /* is implemented ACCURATELY by the vendor-supplied audio-library write,  */
  /* that is sufficient.  Otherwise, requires some kind of IOCTL from here. */
  /* This functionality is IMPORTANT when other realtime I/O is occurring,  */
  /* such as when external MIDI data is being collected from a serial port. */
  /* Since Csound polls for MIDI input at the software synthesis K-rate     */
  /* (the resolution of all software-synthesized events), the user can      */
  /* eliminate MIDI jitter by requesting that both be made synchronous with */
  /* the above audio I/O blocks, i.e. by setting -b to some 1 or 2 K-prds.  */
{
    int             ndx = (int) audioOutData->nrecs % audioOutData->numbuf_out;
    volatile DWORD  *dwflg = &(audioOutData->wavhdr[ndx].dwFlags);
    /*RWD.7.99: only test WHDR_DONE when all blocks have */
    /* been used for the first time! */
    if (audioOutData->nrecs >= audioOutData->numbuf_out) {
      if (audioOutData->rtwin_sleep)            /* if input too fast */
        while ((*dwflg & WHDR_DONE) == 0) Sleep(audioOutData->rtwin_sleep);
      else
        while ((*dwflg & WHDR_DONE) == 0);
      if (audioOutData->rwd_hdrs)
        waveOutUnprepareHeader(audioOutData->outdev,
                               &(audioOutData->wavhdr[ndx]), sizeof(WAVEHDR));
    }
    /* copy data */
    memcpy(audioOutData->wavbuff[ndx], outbuf, nbytes);
    if (audioOutData->rwd_hdrs) {
      audioOutData->wavhdr[ndx].dwBufferLength = nbytes;
      /* RWD.7.99 ...but we do want to do it here */
      waveOutPrepareHeader(audioOutData->outdev,
                           &(audioOutData->wavhdr[ndx]), sizeof(WAVEHDR));
    }
    waveOutWrite(audioOutData->outdev,
                 &(audioOutData->wavhdr[ndx]), sizeof(WAVEHDR));
    /* next buffer */
    audioOutData->nrecs++;
}

/* RWD: extra func for use in cwin_exit(): to close device on premature exit, */
/* and make all this reentry-friendly */

static void rtclose(WAVE_IN_DATA *audioInData, WAVE_OUT_DATA *audioOutData)
{
    /* close output device */
    if (audioOutData != NULL) {
      /* RWD.7.99: just wait for the last block to be DONE, and we're done! */
      /* RWD 12:2001 test we are doing this! */
      int lastblock = (audioOutData->nrecs + (audioOutData->numbuf_out - 1))
                      % audioOutData->numbuf_out;
      volatile DWORD    *dwflg = &(audioOutData->wavhdr[lastblock].dwFlags);
      while ((*dwflg & WHDR_DONE) == 0)
        Sleep(1);
      waveOutReset(audioOutData->outdev);
      waveOutClose(audioOutData->outdev);
      free((void*) audioOutData);
    }
    /* same for input */
    if (audioInData != NULL) {
      waveInStop(audioInData->indev);
      waveInReset(audioInData->indev);
      waveInClose(audioInData->indev);
      free((void*) audioInData);
    }
    /* reset scheduling priority to normal */
    set_priority_level(2);
}

long opcode_count(void *csound)
{
    csound = csound;
    return 0L;
}

void *opcode_init(void *csound)
{
    fprintf(stderr, "Windows MME plugin\n");

    ENGINE->SetPlayopenCallback(csound, playopen);
    ENGINE->SetRtplayCallback(csound, (void (*)(void*, char*, int)) rtplay);
    ENGINE->SetRecopenCallback(csound, recopen);
    ENGINE->SetRtrecordCallback(csound, (int (*)(void*, char*, int)) rtrecord);
    ENGINE->SetRtcloseCallback(csound, (void (*)(void*, void*)) rtclose);

    return NULL;
}

